// This code learns packet rate per source IP per second
// only for traffic going to port 53

#include "agg.hh"

#include <vector>
#include <algorithm>

extern set<string> attackers;

struct restruct
{
  string ip;
  int rate;
};

bool compare(restruct i1, restruct i2)
{
    return (i1.rate > i2.rate);
}
  

int leftr = 0;


void Agg::updateresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
  
  for(int l=0; l<NL; l++)
    {
	double ao = 0;
	
	if (time - resolvers[index][ip].time[l] >= limits[l])
	  {
	    resolvers[index][ip].r[l] = 0;
	    resolvers[index][ip].time[l] = time;
	  }
	resolvers[index][ip].r[l] += req;
    }
  resolvers[index][ip].lasttime = time;
}

void Agg::processresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
  updateresolver(time, ip, req);
}
  
double Agg::shouldfilter(char* ip, long int time, bool instant)
{
  double ratepass = 0;
  
  int index = gettwo(ip);
  
  for(int l=0; l<nlimits; l++)
    {
      double avg;
      double std = 0;
      int samples;
      char choice;
      
      if (time - resolvers[index][ip].time[l] >= limits[l])
	{
	  resolvers[index][ip].r[l] = 0;
	  resolvers[index][ip].time[l] = time;
	}
      
      avg = resolvers[index][ip].avg[l];
      if (resolvers[index][ip].ss[l] > 0)
	std = sqrt(resolvers[index][ip].ss[l]);
      samples = resolvers[index][ip].samples[l];
      
      // Allow for at least one request stdev
      if (std < 1)
	std = 1;
      
      // Allow for larger std given the interval
      if (std < 0.1*limits[l])
	std = 0.1*limits[l];
      
      int allowed = ceil(avg + NUMSTDM*std);
      
      if (resolvers[index][ip].r[l] >= allowed)
	{
	  if (std < 1)
	    {
	      std = 1;	      
	    }	  
	  ratepass += (resolvers[index][ip].r[l] - allowed)/std;
	}
    }
  return ratepass;
}
   
void Agg::loadtraining(string infile)
{
  ifstream in(infile, std::ofstream::in);
  while (in.good())
    {
      char ip[MAXLEN];
      in>>ip;
      if (!in.good())
	break;      
      int index = gettwo(ip);
      auto found = resolvers[index].find(ip);
      record rec;
      resolvers[index][ip] = rec;
      record *r = &resolvers[index][ip];
      
      for (int i=0; i<NL; i++)
	{
	  in>>r->avg[i]>>r->ss[i]>>r->samples[i];
	  r->time[i] = 0;
	  if (i > 0)
	    {
	      if (r->avg[i] < r->avg[i-1])
		r->avg[i] = r->avg[i-1];
	      if (r->ss[i] < r->ss[i-1])
		r->ss[i] = r->ss[i-1];
	    }
	}
      
      resolvers[index][ip].blocked = false;
      resolvers[index][ip].at = 0;
    }
  in.close();
}

int Agg::process(string ip, double outtime, int outlen, bool isquery, char* queryname, bool& isattack, int maxreqs, int ttl)
{
  if (lasttime == 0)
    lasttime = outtime;
  if (outtime - lasttime > 1)
    {
      samples ++;
      double data = apassed + gpassed + amissing + gmissing;
      passed = data;
      double ao = avg;
      avg = avg + (data - avg)/samples;
      ss = ss + (data - ao)*(data - avg);

      passed = apassed + gpassed + amissing + gmissing;
      filtered = afiltered + gfiltered;
      missing = amissing + gmissing;
      
      reqs = 0;
      lasttime = outtime;
    }
  
  char* ips = ip.c_str();
  int index = gettwo(ips);
  if (ip == "")
    return 1;

  // Resolver is missing in training, don't do anything
  if (resolvers[index].find(ip) == resolvers[index].end())
    {
      if (isattack)
	amissing++;
      else
	gmissing++;
      
      passed = apassed + gpassed + amissing + gmissing;
      return 0;
    }


  processresolver(outtime, ips, 1);
  resolvers[index][ips].lasttime = outtime;
  int sf = shouldfilter(ips, outtime, false);
  resolvers[index][ips].at++;

  resolvers[index][ips].xcd *= 0.5;
  if (!noblock && outtime > starttime + 1)
    {
      vector <restruct> tosort;

      for(int i = 0; i < NR; i++)
	{
	  for (auto it = resolvers[i].begin(); it != resolvers[i].end(); it++)
	    {
	      string ipt = it->first;
	      int r = it->second.r[0];
	      restruct rs;
	      rs.ip = ipt;
	      rs.rate = r;
	      tosort.push_back(rs);
	    }
	}
      std::sort(tosort.begin(),tosort.end(), compare);
      int todrop = 0;
      leftr = maxreqs;

      // Block from most to least aggressive until we drop as much traffic as needed
      for(auto it = tosort.begin(); it != tosort.end(); it++)
	{
	  char* ips = it->ip.c_str();
	  int index = gettwo(ips);
	  if (resolvers[index][ips].blocked == false)
	    {
	      todrop += it->rate;
	      resolvers[index][ips].blocked = true;
	      string type;
	      if (attackers.find(ips) != attackers.end())
		type = "attack";
	      else
		type = "good";
	  
	      if (todrop >= leftr)
		{
		  noblock = true;
		  break;
		}
	    }
	}
    }
  if(resolvers[index][ips].blocked == true)
    {
      if (isattack)
	afiltered++;
      else
	gfiltered++;
      filtered = afiltered + gfiltered;
      return 2;
    }
  else
    {
      if (isattack)
	apassed++;
      else
	gpassed++;
      passed = apassed + gpassed;
      return 0;
    }
}

void Agg::reset()
{
  for(int index = 0; index < NR; index++)
    for (auto it = resolvers[index].begin(); it != resolvers[index].end(); it++)
      {
	it->second.blocked = false;
	it->second.xcd = 0;
	it->second.at = 0;
      }
}

void Agg::initFilter(string n, string ty, double t, int nstd, string tf, long s, int nl, int r, double mfp, int m)
{
  delimiters = (int*) malloc(AR_LEN*sizeof(int));

  name = n;
  threshPts = t;
  NUMSTDM = nstd;
  starttime = s;
  nlimits = nl;
  RUN = r;
  infile = tf;
  type = ty;
  fp = mfp;
  mode = m;

  loadtraining(infile);
}

  

