// This code just parses stuff and passes it to other modules

#include "wild.hh"
#include "filter.hh"
#include "hcf.hh"
#include "fq.hh"
//#include "agg.hh" // This can be added if we want to use aggressive sender filter

#include<sys/time.h>
#include<vector>
#include<deque>
#include<algorithm>

vector<Filter*> singles;
vector<Filter*> filters;
vector<Filter*> deployed;
set<string> queries;
deque<int> stats;

double deploymentTime = 0;
int TRIALPERIOD = 1;
int WAIT = 2;
int reqlimit = 0;
bool noblock = false;
bool first = true;
long int starttime = 0;
long int lasttime = 0;
double threshPts;
int nlimits = 9;
int samples = 0;
double avg = 0;
double ss = 0;
double cusum = 0;

int NS = 1;
int RUN = 3600;
int BLOCK = 600;
long int maxreqs = 0;
double mACC = 0;
int reqs = 0;
string readfolder = "";
string infile = "";
string atfile = "";
string extension = "";
int limits[]={1,2,4,8,16,32,64,128,256}; 


map<string,record> resolvers[NR];
set<string> attackers;
long int total = 0;
long int afiltered = 0, apassed = 0;
long int gfiltered = 0, gpassed = 0;
long int passed = 0;

// We store delimiters in this array
int* delimiters;

void updateresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
 
  for(int l=0; l<NL; l++)
    {
      double ao = 0;

      if (time - resolvers[index][ip].time[l] >= limits[l])
        {
	   resolvers[index][ip].r[l] = 0;
	   resolvers[index][ip].time[l] = time;
        }
      resolvers[index][ip].r[l] += req;
    }
  resolvers[index][ip].lasttime = time;
}

void processresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
  updateresolver(time, ip, req);
}

// We can load known attackers from a file as ground truth
void loadattackers(string infile)
{
  ifstream in(infile, std::ofstream::in);
  while (in.good())
    {
      char ip[50];
      int req;
      in>>ip;
      if (!in.good())
	break;
      attackers.insert(ip);
    }
  in.close();
}

// Load training data from a file
void loadtraining(string infile)
{
  ifstream in(infile, std::ofstream::in);
  while (in.good())
    {
      char ip[50];
      in>>ip;
      if (!in.good())
	break;      
      int index = gettwo(ip);
      record rec;
      resolvers[index][ip] = rec;
      record *r = &resolvers[index][ip];
      
      for (int i=0; i<NL; i++)
	{
	  in>>r->avg[i]>>r->ss[i]>>r->samples[i];
	  r->time[i] = 0;
	  if (i > 0)
	    {
	      if (r->avg[i] < r->avg[i-1])
		r->avg[i] = r->avg[i-1];
	      if (r->ss[i] < r->ss[i-1])
		r->ss[i] = r->ss[i-1];
	    }
	}      
      resolvers[index][ip].blocked = false;
      resolvers[index][ip].at = 0;
    }
  in.close();
}

void saveresolvers(string outfile)
{
  // Nothing to save
}

int getavg()
{
  int sum = 0;
  int last = 0;
  if (stats.size() < NS)
    return 0;
  for (auto it = stats.begin(); it != stats.end(); it++)
    {
      sum += *it;
      last = *it;
    }
  return sum/NS;
}

double getover(int max)
{
  int sum = 0;
  int last = 0;
  if (stats.size() < NS)
    return 0;
  for (auto it = stats.begin(); it != stats.end(); it++)
    {
      if (max < *it)
	sum ++;
    }
  return (double)sum/NS;
}

bool deploy(vector<Filter*>* choose, bool isdeployed, double outtime)
{
  // Either rearrange order of deployed and get rid of dups (isdeployed = true)
  // or select candidates to deploy in combo (isdeployed = false)
  int base = 0;
  bool ch = false;
  set<Filter*> combocands;
  set<string> combonames;
  int nf = choose->size();
  int fqs = 0;
  for (auto fit = choose->begin(); fit != choose->end(); fit++)
    {
      double feff = (double)(*fit)->filtered/((*fit)->filtered+(*fit)->passed+1);
      if (isdeployed || (feff >= 0.05 && (combonames.find((*fit)->name) == combonames.end() ||
						  ((*fit)->type == "fqblock"))))
	{
	  combocands.insert(*fit);
	  combonames.insert((*fit)->name);
	  base += (*fit)->filtered;
	  if ((*fit)->type == "fqblock" || (*fit)->type == "fq")
	    fqs ++;
	}
    }
  Filter* toerase = 0;
  if (fqs == 2)
    {
      for (auto fit = combocands.begin(); fit != combocands.end(); fit++)
	if ((*fit)->type == "fqblock" || (*fit)->type == "fqblock-s")
	  {
	    // remove the lower one
	    toerase = *fit;
	  }
      if (toerase != 0)
	combocands.erase(toerase);
    }
  
  // Is the set of candidates enough? is it making a dent?
  long int t = total;
  if (!isdeployed)
    t = apassed + gpassed;
  if (t - base < maxreqs)
    {
      vector<Filter*> savedeployed = deployed;
      deployed.clear();
      double delay = outtime - starttime;
      bool deployedFQ = false;
      int nfilter = 0;
      bool enough = false;
      for (auto fit = combocands.begin(); fit != combocands.end(); fit++)
	{
	  if ((*fit)->type == "ur" || (*fit)->type == "ur-s")
	    {
	      deployed.push_back(*fit);
	      nfilter += (*fit)->filtered;
	    }
	}
      if (total - nfilter < maxreqs)
	enough = true;
      for (auto fit = combocands.begin(); fit != combocands.end(); fit++)
	{
	  if (((*fit)->type == "hcf" || (*fit)->type == "hcf-s") && !enough)
	    {
	      deployed.push_back(*fit);
	      nfilter += (*fit)->filtered;
	    }
	}
      if (total - nfilter < maxreqs)
	enough = true;

      for (auto fit = combocands.begin(); fit != combocands.end(); fit++)
	{
	  if (((*fit)->type == "fq" || (*fit)->type == "fqblock" || (*fit)->type == "fq-s" || (*fit)->type == "fqblock-s") && !enough)
	    {
	      deployed.push_back(*fit);
	      nfilter += (*fit)->filtered;
	      deployedFQ = true;
	    }
	}
      if (total - nfilter < maxreqs)
	enough = true;
      for (auto fit = combocands.begin(); fit != combocands.end(); fit++)
	{
	  if (((*fit)->type == "wild" || (*fit)->type == "wild-s" )  && !enough)
	    {
	      deployed.push_back(*fit);	      
	    }
	}
      for (auto fit = deployed.begin(); fit != deployed.end(); fit++)
	if (std::find(savedeployed.begin(), savedeployed.end(), *fit)
		  == savedeployed.end())
	  ch = true;
      for (auto fit = savedeployed.begin(); fit != savedeployed.end(); fit++)
	if (std::find(deployed.begin(), deployed.end(), *fit)
		  == deployed.end())
	  ch = true;
    }
  return ch;
}

int process(char* buffer, double &outtime, int& outlen, int& outttl)
{
  string ip = "";

  int todrop = 0;
  bool isquery;
  char queryname[MAXLEN];
  bool toprocess = shouldprocess2(buffer, outtime, outlen, delimiters,
				  ip, starttime, RUN, isquery, queryname, outttl);

  if (outtime > starttime + RUN)
    {
      for (auto fit = filters.begin(); fit != filters.end(); fit++)
	{
	  if ((*fit)->type == "fqblock")
	    (*fit)->printtable();
	}
      exit(0);
    }
  if (outtime - lasttime > 1)
    {
      total = apassed + gpassed + afiltered + gfiltered;
      if (outtime < starttime)
	{
	  samples ++;	  
	  double data = apassed + gpassed;
	  double ao = avg;
	  avg = avg + (data - avg)/samples;
	  ss = ss + (data - ao)*(data - avg);
	  
	  double stdev = 0;
	  if (samples > 1)
	    stdev = sqrt(ss/(samples-1));
	  maxreqs = mACC*avg;
	  if (maxreqs == 0)
	    maxreqs = reqlimit;
	}
      if (deploymentTime + TRIALPERIOD < outtime && deploymentTime > 0)
	{
	  stats.push_back(apassed+gpassed);
	}
      if (stats.size() > NS)
	stats.pop_front();
      passed = passed*0.8 + (apassed + gpassed)*0.2;
      cout<<std::fixed;      
      cout<<outtime<<" Attack passed "<<apassed<<" filtered "<<afiltered<<endl;
      cout<<outtime<<" Good passed "<<gpassed<<" filtered "<<gfiltered<<endl;
      cout<<outtime<<" Total passed "<<gpassed+apassed<<" smoothed "<<passed<<" filtered "<<gfiltered+afiltered<<" reqs "<<reqs<<" Max "<<maxreqs<<" getavg "<<getavg()<<" getover "<<getover(maxreqs)<<noblock<<endl;

      if (!noblock && outtime > starttime && apassed + gpassed < maxreqs*0.9)
	{
	  for (auto fit = filters.begin(); fit != filters.end(); fit++)
	    {
	      if (((*fit)->type == "wild") && (*fit)->noblock == false)
		{
		  (*fit)->setnoblock(true);
		}
	    }
	  for (auto fit = singles.begin(); fit != singles.end(); fit++)
	    {
	      if (((*fit)->type == "wild-s") && (*fit)->noblock == false)
		{
		  (*fit)->setnoblock(true);
		}
	    }
	  for (auto fit = singles.begin(); fit != singles.end(); fit++) // This is only important for aggressive sender filtering
	    {
	      if (((*fit)->type == "agg-s" || (*fit)->type == "agg") && (*fit)->noblock == false)
		{
		  (*fit)->setnoblock(true);
		}
	    }
	  noblock = true;
	  passed = apassed+gpassed;
	}
      // Select new filters
      else if (noblock && (getover(maxreqs) == 1 || deploymentTime == 0) && maxreqs > 0 && outtime > starttime &&
	       (deploymentTime == 0 || outtime - deploymentTime > TRIALPERIOD))
	{
	  noblock = false;
	  double difft = outtime - deploymentTime;
	  for (auto fit = filters.begin(); fit != filters.end(); fit++)
	    {
	      if ((*fit)->type == "wild")
		{
		  (*fit)->reset();
		  (*fit)->setnoblock(false);
		}
	    }
	  for (auto fit = singles.begin(); fit != singles.end(); fit++)
	    {
	      if ((*fit)->type == "wild-s")
		{
		  (*fit)->reset();
		  (*fit)->setnoblock(false);
		}
	    }
	}
      // This is only important for aggressive sender filtering
      if (apassed + gpassed > maxreqs)
	{
	  for (auto fit = singles.begin(); fit != singles.end(); fit++)
	    {
	      if ((*fit)->type == "agg-s")
		{
		  (*fit)->setnoblock(false);
		}
	    }
	}
      if (outtime > starttime && total > maxreqs)
	{
	  bool change = false;
	  set<Filter*> cands;

	  for (auto fit = singles.begin(); fit != singles.end(); fit++)
	    {
	      (*fit)->getfeff();
	      double feff = (double)(*fit)->filtered/((*fit)->filtered+(*fit)->passed+1);
	      if ((*fit)->passed < maxreqs && (*fit)->passed > 0 && feff > 0.1) 
		{ 
		  (*fit)->iscand = (*fit)->iscand*0.5 + 0.5;
		  cout<<std::fixed;
		  cands.insert(*fit);
		}
	      else
		{
		  cout<<std::fixed;
		  (*fit)->iscand = (*fit)->iscand*0.5;
		}
	    }
	  if (cands.size() == 0 && singles.size() == 1 && outtime > starttime)
	    {
	      for (auto fit = singles.begin(); fit != singles.end(); fit++)
		{
		  if (std::find(deployed.begin(), deployed.end(), *fit)
		      == deployed.end())
		    {
		      deployed.push_back(*fit);
		      if (first)
			deploymentTime = outtime;
		      first = false;
		    }
		}
	    }
	
	  double bestfp = 100;
	  double bestiscand = 0;
	  Filter* best = 0;
	  for (auto cit = cands.begin(); cit != cands.end(); cit++)
	    {
	      // The candidate has low false positives, it filters actively and the attack is ongoing
	      if (((*cit)->fp < bestfp || ((*cit)->fp == bestfp && (*cit)->iscand > bestiscand)) && gfiltered + afiltered + gpassed + apassed > maxreqs)
		{
		  best = *cit;
		  bestfp = (*cit)->fp;
		  bestiscand = (*cit)->iscand;
		  double diff = outtime - deploymentTime;
		}
	    }
	  bool isdeployed = false;
	  if (best && outtime - deploymentTime > TRIALPERIOD)
	    {
	      for (auto dit = deployed.begin(); dit != deployed.end(); dit ++)
		{
		  if ((*dit)->name == best->name)
		    {
		      isdeployed = true;
		      break;
		    }
		}
	      if (isdeployed == false)
		{
		  change = true;
		}
	    }
	  /* Put this back for aggressive sender filter
	  for (auto fit = singles.begin(); fit != singles.end(); fit++)
	    {
	      if ((*fit)->type == "agg-s" || (*fit)->type == "agg")
		{
		  cout<<"Setting apassed "<<apassed<<" gpassed "<<gpassed<<endl;
		  (*fit)->setpassed(apassed, gpassed);
		}
	    }
	  */
	  if (change || ((passed > maxreqs || (apassed+gpassed > maxreqs && deploymentTime == 0)) && noblock == false))
	    {
	      if (best != 0)
		{
		  if (std::find(deployed.begin(), deployed.end(), best) == deployed.end())
		    {
		      deployed.clear();
		      deployed.push_back(best);
		      deploy(&deployed, true, outtime);
		      deploymentTime = outtime;
		    }
		}
	      else
		{
		  double delay = outtime - starttime;
		  
		  int bestpassed = 0;
		  Filter* best = 0;
		  // Find filter candidates
		  bool ch = deploy(&filters, false, outtime);
		  if (ch)
		    deploymentTime = outtime;
		}
	    }
	  
	  int nf = deployed.size();
	  bool erased = false;
	  do
	    {
	      erased = false;
	      int toerase = -1;
	      int pos = 0;
	      // If there is a very effective filter, remove the others                                                                                                                               
              double maxeff = 0.8, eff = 0;
              for (auto dit = deployed.begin(); dit != deployed.end(); dit++)
		{
                  double ceff = (double)(*dit)->filtered/((*dit)->filtered + (*dit)->passed + 1);
                  if (ceff > eff)
                    eff = ceff;
                }
              cout<<"Max eff "<<eff<<endl;
              for (auto dit = deployed.begin(); dit != deployed.end(); dit++)
                {
                  double ceff = (double)(*dit)->filtered/((*dit)->filtered + (*dit)->passed + 1);
                  cout<<"Ceff "<<ceff<<" eff "<<eff<<" maxeff "<<maxeff<<" size "<<deployed.size()<<endl;
                  if (!(ceff < eff && eff > maxeff && deployed.size() > 1))
		    {
		      continue;
		    }
		  else
		    {
		      toerase = pos;
		      break;
		    }
		  pos++;
		}
	      if (toerase > -1)
		{	       
		  deployed.erase(deployed.begin()+toerase);
		  erased = true;
		}
	    }while(erased);
	}
      for (auto fit = filters.begin(); fit != filters.end(); fit++)
	(*fit)->resetcounts();

      for (auto fit = singles.begin(); fit != singles.end(); fit++)
	(*fit)->resetcounts();

      todrop = apassed + gpassed - maxreqs;
      apassed = afiltered = 0;
      gpassed = gfiltered = 0;
      reqs = 0;
      lasttime = outtime;

      cout<<std::fixed<<outtime<<" deployed ";
      for (auto dit = deployed.begin(); dit != deployed.end(); dit++)
	cout<<(*dit)->name<<"-"<<(*dit)->type<<" ";
      cout<<std::fixed<<" deploymentTime "<<deploymentTime<<endl;
    }

  if (ip == "")
    return 1;

  
  if (outtime < starttime)
    {
      if (attackers.find(ip) != attackers.end())
	apassed++;
      else       
	gpassed++;       
      return 1;
    }

    if (!toprocess)
    {
      return 1;
    }

    // Apply ground truth detection
    bool isattack = false;
    if(!isquery || outlen > 256)
      {
	// Everything not query is attack
	isattack = true;
      }
    // Everything fitting specific query name is attack
    for (auto qit = queries.begin(); qit != queries.end(); qit++)
      if (strstr(queryname, qit->c_str()) != 0)
	{
	  isattack = true;	  
	}
    if (attackers.find(ip) != attackers.end())
      {
	isattack = true;
      }

    int totaloutcome = 0;
    for (auto dit = deployed.begin(); dit != deployed.end(); dit++)
      {
	int outcome;
	outcome = (*dit)->process(ip, outtime, outlen, isquery, queryname, isattack, maxreqs, outttl);
	if (outcome == 2)
	  {
	    totaloutcome = 2;
	    break;
	  }
      }
    if (totaloutcome == 0)
      {
	if (isattack)
	  apassed++;
	else
	  gpassed++;
	for (auto fit = filters.begin(); fit != filters.end(); fit++)
	  {
	    if (std::find(deployed.begin(), deployed.end(), *fit) == deployed.end())
	      {
		int outcome = (*fit)->process(ip, outtime, outlen, isquery, queryname, isattack, maxreqs, outttl);
	      }
	  }
      }
    else
      {
	if (isattack)
	  afiltered++;
	else
	  gfiltered++;
      }
    for (auto fit = singles.begin(); fit != singles.end(); fit++)
      {
	if (std::find(deployed.begin(), deployed.end(), *fit) == deployed.end())
	  {
	    int todrop = apassed + gpassed - maxreqs;
	    if (todrop < 0)
	      todrop = 0;
	    int outcome = (*fit)->process(ip, outtime, outlen, isquery, queryname, isattack, maxreqs, outttl);
	  }
      }
    return 0;
}


void printHelp()
{
  printf ("drop-train\n(C) 2018 University of Southern California.\n\n");

  printf ("-h                             Print this help\n");
  printf ("-r <folder>                    Folder with pcap.xz files to use in testing\n");
  printf ("-s <epoch>                     Start processing from this epoch time in UTC\n");
  printf ("-b <secs>                      Length of deciding what to block, stop after this time period\n");
  printf ("-m <secs>                      Measure for this much longer, after stopping the decision process\n");
  printf ("-e <ext>                       Only process files with this extension in the name (e.g., lax, mia)\n");
  printf ("-n <numwin>                    Use this many time windows for training, max 9\n");
  printf ("-f <file>                      Read training output from this file\n");
  printf ("-a <file>                      Optionally read attack IPs from this file\n");
  printf ("-T <pts>                       Block threshold as floating point number\n");
  printf ("-l <reqs>                      Specify max requests per second\n");
}


// This is deployment version, we load trained values
// and use them to block resolvers that are more aggressive than
// their model

int main(int argc, char** argv)
{
  delimiters = (int*) malloc(AR_LEN*sizeof(int));
  char c;
  set<string> wildtrain;
  set<string> HCFtrain;
  set<string> URtrain;
  set<string> FQtrain;

  for (int i = 0; i<argc; i++)
    cout<<argv[i]<<" ";
  cout<<endl;
  while ((c = getopt (argc, argv, "hs:n:m:b:e:r:f:T:a:l:d:w:q:H:u:F:")) != '?')
    {
      if ((c == 255) || (c == -1))
	break;

      switch (c)
	{
	case 'h':
	  printHelp ();
	  return (0);
	  break;
	case 'T':
	  threshPts = atof(optarg);
	  break;
	case 'r':
	  readfolder = optarg;
	  break;
	case 'l':
	  reqlimit = atoi(optarg);
	  break;
	case 'w':
	  infile = optarg;
	  wildtrain.insert(infile);
	  break;
	case 'F':
	  infile = optarg;
	  FQtrain.insert(infile);
	  break;
	case 'H':
	  infile = optarg;
	  HCFtrain.insert(infile);
	  break;
	case 'u':
	  infile = optarg;
	  URtrain.insert(infile);
	  break;
	case 'q':
	  queries.insert(optarg);
	  break;
	case 'a':
	  atfile = optarg;
	  break;
	case 's':
	  starttime = atol(optarg);
	  break;
	case 'n':
	  nlimits = atoi(optarg);
	  if (nlimits > NL)
	    {
	      cout<<"Maximum limits supported "<<NL<<endl;
	      exit(0);
	    }
	  break;
	case 'b':
	  mACC = atof(optarg); // how many times an avg do we want to let through
	  break;
	case 'm':
	  RUN = atoi(optarg); // how long to run
	  break;
	case 'e':
	  extension = optarg;
	  break;
	default:
	  break;
	}
    }
  cout<<"Start "<<starttime<<" dir "<<readfolder<<" nlimits "<<nlimits<<" run for "<<RUN<<" block for "<<BLOCK<<" threshPts "<<threshPts<<" maxreqs "<<maxreqs<<endl;
  if (readfolder == "")
    {
      cout<<"You must specify a directory with pcap.xz files\n";
      exit(0);
    }
  if (infile == "")
    {
      cout<<"You must specify the file with training output\n";
      exit(0);
    }
  // Assume the attack has started and this is why we're being invoked
  // If we are testing it helps to use a file with attackers
  if (atfile != "")
    loadattackers(atfile);

  for (auto fit=FQtrain.begin(); fit != FQtrain.end(); fit++)
    {
      FQ* f = new FQ();
      f->setlearn(10000);
      f->initFilter((*fit), "fq", 0.3, 0, (*fit), starttime, 0, RUN, 0);
      filters.push_back(f);
      FQ* fb = new FQ();
      fb->setlearn(10000);
      fb->initFilter((*fit), "fqblock", 0.3, 0, (*fit), starttime, 0, RUN, 0);
      fb->setblock(true);
      filters.push_back(fb);
      f = new FQ();
      f->setlearn(10000);
      f->initFilter((*fit), "fq-s", 0.3, 0, (*fit), starttime, 0, RUN, 0);
      singles.push_back(f);
      fb = new FQ();
      fb->setlearn(10000);
      fb->initFilter((*fit), "fqblock-s", 0.3, 0, (*fit), starttime, 0, RUN, 0);
      fb->setblock(true);
      singles.push_back(fb);
    }
 
  for (auto wit=wildtrain.begin(); wit != wildtrain.end(); wit++)
    {
      Wild* w = new Wild();
      w->initFilter((*wit), "wild", threshPts, 5, (*wit), starttime, nlimits, RUN, 0.05);
      filters.push_back(w);
      w = new Wild();
      w->initFilter((*wit), "wild-s", threshPts, 5, (*wit), starttime, nlimits, RUN, 0.05);
      singles.push_back(w);
    }

  /*
  // Include only if we are doing aggressive sender filtering
    for (auto wit=wildtrain.begin(); wit != wildtrain.end(); wit++)
    {
      Agg* w = new Agg();
      
      //w->initFilter((*wit), "agg", threshPts, 5, (*wit), starttime, nlimits, RUN, 0.05);
      //filters.push_back(w);
      //w = new Agg();
      w->initFilter((*wit), "agg-s", threshPts, 5, (*wit), starttime, nlimits, RUN, 0.05);
      singles.push_back(w);
    } 
  */    

  for (auto hit=HCFtrain.begin(); hit != HCFtrain.end(); hit++)
    {
      HCF* h = new HCF();
      h->setmode(1);
      h->initFilter((*hit), "hcf", 0, 0, (*hit), starttime, 0, RUN, 0.02);
      filters.push_back(h);
      h = new HCF();
      h->setmode(1);
      h->initFilter((*hit), "hcf-s", 0, 0, (*hit), starttime, 0, RUN, 0.02);
      singles.push_back(h);
    }

  for (auto uit=URtrain.begin(); uit != URtrain.end(); uit++)
    {
      HCF* u = new HCF();
      u->setmode(2);
      u->initFilter((*uit), "ur", 0, 0, (*uit), starttime, 0, RUN, 0.013);
      filters.push_back(u);
      u = new HCF();
      u->setmode(2);
      u->initFilter((*uit), "ur-s", 0, 0, (*uit), starttime, 0, RUN, 0.013);
      singles.push_back(u);
    }

  // Use for reading txt files that are synthetically generated
  //readfiles(readfolder.c_str(), process, saveresolvers, "",
  //      extension, starttime, RUN);
  
  //Use for reading pcap files
  loadfiles(readfolder.c_str(), process, saveresolvers, "",
 	    extension, starttime, RUN);
}
  

