// This code learns packet rate per source IP per several time windows
// only for traffic going to port 53

#include "wild.hh"


void Wild::updateresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
  
  for(int l=0; l<NL; l++)
    {
	double ao = 0;
	
	if (time - resolvers[index][ip].time[l] >= limits[l])
	  {
	    resolvers[index][ip].r[l] = 0;
	    resolvers[index][ip].time[l] = time;
	  }
	resolvers[index][ip].r[l] += req;
    }
  resolvers[index][ip].lasttime = time;
}

void Wild::processresolver(long time, char* ip, int req)
{
  int index = gettwo(ip);
  updateresolver(time, ip, req);
}
  
double Wild::shouldfilter(char* ip, long int time, bool instant)
{
  double ratepass = 0;
  
  int index = gettwo(ip);
  
  for(int l=0; l<nlimits; l++)
    {
      double avg;
      double std = 0;
      int samples;
      char choice;
      
      if (time - resolvers[index][ip].time[l] >= limits[l])
	{
	  resolvers[index][ip].r[l] = 0;
	  resolvers[index][ip].time[l] = time;
	}
      
      avg = resolvers[index][ip].avg[l];
      if (resolvers[index][ip].ss[l] > 0)
	std = sqrt(resolvers[index][ip].ss[l]);
      samples = resolvers[index][ip].samples[l];
      
      // Allow for at least one request stdev
      if (std < 1)
	std = 1;
      
      // Allow for larger std given the interval
      if (std < 0.1*limits[l])
	std = 0.1*limits[l];
      
      int allowed = ceil(avg + NUMSTDM*std);
      
      if (resolvers[index][ip].r[l] >= allowed)
	{
	  if (std < 1)
	    {
	      std = 1;	      
	    }	  
	  ratepass += (resolvers[index][ip].r[l] - allowed)/std;
	}
    }
  return ratepass;
}
   
void Wild::loadtraining(string infile)
{
  ifstream in(infile, std::ofstream::in);
  while (in.good())
    {
      char ip[MAXLEN];
      in>>ip;
      if (!in.good())
	break;      
      int index = gettwo(ip);
      auto found = resolvers[index].find(ip);
      record rec;
      resolvers[index][ip] = rec;
      record *r = &resolvers[index][ip];
      
      for (int i=0; i<NL; i++)
	{
	  in>>r->avg[i]>>r->ss[i]>>r->samples[i];
	  r->time[i] = 0;
	  if (i > 0)
	    {
	      if (r->avg[i] < r->avg[i-1])
		r->avg[i] = r->avg[i-1];
	      if (r->ss[i] < r->ss[i-1])
		r->ss[i] = r->ss[i-1];
	    }
	}
      
      resolvers[index][ip].blocked = false;
      resolvers[index][ip].at = 0;
    }
  in.close();
}

int Wild::process(string ip, double outtime, int outlen, bool isquery, char* queryname, bool& isattack, int maxreqs, int ttl)
{
  if (lasttime == 0)
    lasttime = outtime;
  if (outtime - lasttime > 1)
    {
      samples ++;
      double data = apassed + gpassed + amissing + gmissing;
      passed = data;
      double ao = avg;
      avg = avg + (data - avg)/samples;
      ss = ss + (data - ao)*(data - avg);


      passed = apassed + gpassed + amissing + gmissing;
      filtered = afiltered + gfiltered;
      missing = amissing + gmissing;
      
      reqs = 0;
      lasttime = outtime;
    }
  
  char* ips = ip.c_str();
  int index = gettwo(ips);
  if (ip == "")
    return 1;

  // Resolver is missing in training, don't do anything
  if (resolvers[index].find(ip) == resolvers[index].end())
    {
      if (isattack)
	amissing++;
      else
	gmissing++;
      passed = apassed + gpassed + amissing + gmissing;
      return 0;
    }


  processresolver(outtime, ips, 1);
  resolvers[index][ips].lasttime = outtime;
  int sf = shouldfilter(ips, outtime, false);
  resolvers[index][ips].at++;

  resolvers[index][ips].xcd *= 0.5;
  if (sf > 0)
    {
      string type;
      if (isattack)
	type = "attack";
      else
	type = "good";
      resolvers[index][ips].xcd += sf;
      double xr = resolvers[index][ips].xcd/resolvers[index][ips].at;

      
      if (resolvers[index][ips].blocked == false && resolvers[index][ips].at/resolvers[index][ips].r[0]  >= ASAMPLES && xr >= threshPts && !noblock) 
	resolvers[index][ips].blocked = true;
    }

  if(resolvers[index][ips].blocked == true)
    {
      if (isattack)
	afiltered++;
      else
	gfiltered++;
      filtered = afiltered + gfiltered;
      return 2;
    }
  else
    {
      if (isattack)
	apassed++;
      else
	gpassed++;
      passed = apassed + gpassed;
      return 0;
    }
}

void Wild::reset()
{
  for(int index = 0; index < NR; index++)
    for (auto it = resolvers[index].begin(); it != resolvers[index].end(); it++)
      {
	it->second.blocked = false;
	it->second.xcd = 0;
	it->second.at = 0;
      }
}

void Wild::initFilter(string n, string ty, double t, int nstd, string tf, long s, int nl, int r, double mfp)
{
  delimiters = (int*) malloc(AR_LEN*sizeof(int));

  name = n;
  threshPts = t;
  NUMSTDM = nstd;
  starttime = s;
  nlimits = nl;
  RUN = r;
  infile = tf;
  type = ty;
  fp = mfp;

  loadtraining(infile);
}

  

